/*
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 * Copyright (C) September, 2002 Matsushita Electric Industrial Co., Ltd.
 *	Takahiko Suzuki <tsuzuki@vrl.mei.co.jp>
 * Copyright (C) October, 2002 Matsushita Electric Industrial Co., Ltd.
 *	Takahiko Suzuki <tsuzuki@vrl.mei.co.jp>
 * 
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef  DHCP_UPDATE
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#endif /* defined(DHCP_UPDATE) */

#ifndef DHCP_UPDATE
#include "dhcpcd.h"
#endif /* !defined(DHCP_UPDATE) */

#include "client.h"
#include "signals.h"

#ifdef  DHCP_UPDATE
#include "pathnames.h"
#endif /* defined(DHCP_UPDATE) */

volatile int	giWhile;

#ifdef	DHCP_TIME_MODIFY	// A.Tsuji 2001.07.09
extern time_t dhcpUptime(long * pmsec);
#endif

struct in_addr  inform_ipaddr;
char		*ProgramName	=	NULL;
char            **ProgramEnviron=       NULL;
char		*IfName		=	DEFAULT_IFNAME;
int		IfName_len	=	DEFAULT_IFNAME_LEN;
char		*HostName	=	NULL;
int		HostName_len	=	0;
char		*Cfilename	=	NULL;
unsigned char	*ClassID	=	NULL;
int		ClassID_len	=	0;
unsigned char	*ClientID	=	NULL;
int		ClientID_len	=	0;
void		*(*currState)()	=	&dhcpReboot;
int		DebugFlag	=	0;
int		BeRFC1541	=	0;
unsigned	LeaseTime	=	DEFAULT_LEASETIME;
int		ReplResolvConf	=	1;
int		SetDomainName	=	0;
int		SetHostName	=	0;
int             BroadcastResp   =       0;
time_t          TimeOut         =	DEFAULT_TIMEOUT;
int 		magic_cookie    =       0;
unsigned short  dhcpMsgSize     =       0;
unsigned        nleaseTime      =       0;
int		DoCheckSum	=	0;
int		TestCase	=	0;
int		Window		=	0;

#ifdef  DHCP_UPDATE
char            *IfNameExt      =       DEFAULT_IFNAME;
int             IfNameExt_len   =       DEFAULT_IFNAME_LEN;
int             ReplNISConf     =       1;
int             ReplNTPConf     =       1;
char            *ConfigDir      =       CONFIG_DIR;
#endif /* defined(DHCP_UPDATE) */

/*****************************************************************************/
void print_version()
{
#ifndef	DHCP_MODIFY
  fprintf(stderr,"\
DHCP Client Daemon v."VERSION"\n\
Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>\n\
Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>\n\
Location: http://www.phystech.com/download/\n");
#else	// DHCP_MODIFY
  fprintf(stderr, "%s%s%s%s%s",
	  "DHCP Client Daemon v."VERSION"\n",
	  "Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>\n",
	  "Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>\n",
	  "Location: http://www.phystech.com/download/\n",
	  "Modified by Matsushita Electric Industrial Co., Ltd.\n");
#endif	// DHCP_MODIFY
}
/*****************************************************************************/
int main(argn,argc,argv)
int argn;
char *argc[],*argv[];
{
  int killFlag		=	0;
  int versionFlag	=	0;
  int s			=	1;
  int k			=	1;
  int i			=	1;

#ifdef  DHCP_UPDATE
  int j;

/*
 * Ensure that fds 0, 1, 2 are open, to /dev/null if nowhere else.
 * This way we can close 0, 1, 2 after forking the daemon without clobbering
 * a fd that we are using (such as our sockets). This is necessary if
 * this program is run from init scripts where 0, 1, and/or 2 may be closed.
 */
 j=open("/dev/null",O_RDWR);
 while ( j < 2 && j >= 0 ) j = dup(j);
 if ( j > 2 ) close(j);
#endif  /* defined(DHCP_UPDATE) */

#if 0
  {
    int iCounter = 0;

    giWhile = 0;
    while(!giWhile)
    {
      iCounter ++;
    }
  }
#endif

#ifdef	DHCP_TIME_MODIFY
  {
    long secs, msecs;

    secs = dhcpUptime(&msecs);
    printf( "DHCPcd: start. [%d.%d]\n", secs, msecs);	/* syslog isn't opened. */
  }
#endif

  if ( getuid() )
    {
      fprintf(stderr,"****  %s: not a superuser\n",argc[0]);
      exit(1);
    }

  while ( argc[i] )
    if ( argc[i][0]=='-' )
prgs: switch ( argc[i][s] )
	{
	  case 0:
	    i++;
	    s=1;
	    break;
	  case 'k':
	    s++;
	    killFlag=SIGHUP;
	    goto prgs;
	  case 'n':
	    s++;
	    killFlag=SIGALRM;
	    goto prgs;
	  case 'd':
	    s++;
	    DebugFlag=1;
	    goto prgs;
	  case 'r':
	    s++;
	    BeRFC1541=1;
	    goto prgs;
	  case 'D':
	    s++;
	    SetDomainName=1;
	    goto prgs;
	  case 'H':
	    s++;
	    SetHostName=1;
	    goto prgs;
	  case 'R':
	    s++;
	    ReplResolvConf=0;
	    goto prgs;
#ifdef  DHCP_UPDATE
	  case 'Y':
	    s++;
	    ReplNISConf=0;
	    goto prgs;
	  case 'N':
	    s++;
	    ReplNTPConf=0;
	    goto prgs;
#endif /* defined(DHCP_UPDATE) */
	  case 'V':
	    s++;
	    versionFlag=1;
	    goto prgs;
	  case 'c':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    Cfilename=argc[i++];
	    if ( Cfilename == NULL || Cfilename[0] == '-' ) goto usage;
	    s=1;
	    break;
#ifdef  DHCP_UPDATE
	  case 'L':
	    if ( argc[i][s+1] ) goto usage;
	    i++;
	    ConfigDir=argc[i++];
	    if ( ConfigDir == NULL || ConfigDir[0] != '/' ) goto usage;
	    s=1;
	    break;
#endif /* defined(DHCP_UPDATE) */
	  case 'i':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    ClassID=argc[i++];
	    if ( ClassID == NULL || ClassID[0] == '-' ) goto usage;
	    s=1;
	    if ( (ClassID_len=strlen(ClassID)) < CLASS_ID_MAX_LEN+1 ) break;
	    fprintf(stderr,"****  %s: too long ClassID string: strlen=%d\n",
	    argc[0],ClassID_len);
	    goto usage;
	  case 'I':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    ClientID=argc[i++];
	    if ( ClientID == NULL || ClientID[0] == '-' ) goto usage;
	    s=1;
	    if ( (ClientID_len=strlen(ClientID)) < CLIENT_ID_MAX_LEN+1 ) break;
	    fprintf(stderr,"****  %s: too long ClientID string: strlen=%d\n",
	    argc[0],ClientID_len);
	    goto usage;
	  case 'h':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    HostName=argc[i++];
	    if ( HostName == NULL || HostName[0] == '-' ) goto usage;
	    s=1;
	    if ( (HostName_len=strlen(HostName)+1) < HOSTNAME_MAX_LEN ) break;
	    fprintf(stderr,"****  %s: too long HostName string: strlen=%d\n",
	    argc[0],HostName_len);
	    break;
	  case 't':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    if ( argc[i] )
	      TimeOut=atol(argc[i++]);
	    else
	      goto usage;
	    s=1;
	    if ( TimeOut > 0 ) break;
	    goto usage;
	  case 'w':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
            i++;
	    if ( argc[i] )
	      Window=atol(argc[i++]);
            else
	      goto usage;
	    s=1;
	    if ( Window >= 0 ) break;
	    goto usage;
	  case 's':
	    if ( argc[i][s+1] ) goto usage;
	    i++;
	    if ( argc[i] && inet_aton(argc[i],&inform_ipaddr) )
	      i++;
	    else
	      memset(&inform_ipaddr,0,sizeof(inform_ipaddr));
	    currState = &dhcpInform;
	    s=1;
	    break;
	  case 'B':
	    s++;
	    BroadcastResp=1;
	    goto prgs;
	  case 'C':
	    s++;
	    DoCheckSum=1;
	    goto prgs;
	  case 'T':
	    s++;
	    TestCase=1;
	    goto prgs;
	  case 'l':
#ifdef  DHCP_UPDATE
	    if ( argc[i][s+1] ) goto usage;
#endif /* defined(DHCP_UPDATE) */
	    i++;
	    if ( argc[i] )
	      LeaseTime=atol(argc[i++]);
	    else
	      goto usage;
	    s=1;
	    if ( LeaseTime > 0 ) break;
          default:
usage:	    print_version();
	    fprintf(stderr,
#ifndef DHCP_UPDATE
"Usage: dhcpcd [-dknrBCDHRT] [-l leasetime] [-h hostname] [-t timeout]\n\
       [-i vendorClassID] [-I ClientID] [-c filename] [-s [ipaddr]]\n\
       [-w windowsize] [interface]\n");
#else  /* defined(DHCP_UPDATE) */
"Usage: dhcpcd [-dknrBCDHNRTY] [-l leasetime] [-h hostname] [-t timeout]\n\
       [-i vendorClassID] [-I ClientID] [-c filename] [-s [ipaddr]]\n\
       [-w windowsize] [-L ConfigDir] [interface]\n");
#endif /* !defined(DHCP_UPDATE) */
	    exit(1);
	}
    else
      argc[k++]=argc[i++];
  if ( k > 1 )
    {

#ifndef DHCP_UPDATE
      if ( (IfName_len=strlen(argc[1])) > IFNAMSIZ )
        goto usage;
      else
        IfName=argc[1];
#else  /* defined(DHCP_UPDATE) */
      if ( (IfNameExt_len=strlen(argc[1])) > IFNAMSIZ ) goto usage;
      IfNameExt=argc[1];
      IfName=IfNameExt;
      IfName_len=IfNameExt_len;
      s=0;
      while ( IfNameExt[s] )
	if ( IfNameExt[s] == ':' )
	  {
	    IfName=(char *)malloc(s+1);
	    memcpy(IfName,IfNameExt,s);
	    IfName[s]=0;
	    IfName_len=s;
	    break;
	  }
	else
	  s++;
#endif /* !defined(DHCP_UPDATE) */

    }
  ProgramName=argc[0];
  ProgramEnviron=argv;

#ifdef  DHCP_UPDATE
  umask(022);
#endif /* defined(DHCP_UPDATE) */

  if ( killFlag ) killPid(killFlag);

#ifndef DHCP_UPDATE
  checkIfAlreadyRunning();
#else  /* defined(DHCP_UPDATE) */
  if ( ! TestCase ) checkIfAlreadyRunning();
#endif /* !defined(DHCP_UPDATE) */

  if ( versionFlag ) print_version();
  openlog(PROGRAM_NAME,LOG_PID|LOG_CONS,LOG_LOCAL0);
  signalSetup();

#ifdef  DHCP_UPDATE
  if ( mkdir(ConfigDir,S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) && errno != EEXIST )
    {
      syslog(LOG_ERR,"mkdir(\"%s\",0): %m\n",ConfigDir);
      exit(1);
    }
#endif /* defined(DHCP_UPDATE) */

  magic_cookie = htonl(MAGIC_COOKIE);
  dhcpMsgSize = htons(sizeof(dhcpMessage));
  nleaseTime = htonl(LeaseTime);
#ifdef	DHCP_DEBUG_INFO
  {
    long secs, msecs;

    secs = dhcpUptime(&msecs);
    syslog( LOG_INFO, "DHCP transition start. [%d.%d]\n", secs, msecs );
  }
#endif
  alarm(TimeOut);
  do {

#ifdef	DHCP_DEBUG_INFO
    if ( DebugFlag ) {
#ifndef	DHCP_TIME_MODIFY
    //      syslog(LOG_ERR, "Transition of the state... [%d]", time(NULL));
      if ( currState == &dhcpReboot ) {
	syslog(LOG_DEBUG, " [DHCP_REBOOT] -> \n");
      } else if ( currState == &dhcpInform ) {
	syslog(LOG_DEBUG, " [DHCP_INFORM] -> \n");
      } else if ( currState == &dhcpBound ) {
	syslog(LOG_DEBUG, " [DHCP_BOUND] -> \n");
      } else if ( currState == &dhcpRenew ) {
	syslog(LOG_DEBUG, " [DHCP_RENEW] -> \n");
      } else if ( currState == &dhcpRebind ) {
	syslog(LOG_DEBUG, " [DHCP_REBIND] -> \n");
      } else if ( currState == &dhcpInit ) {
	syslog(LOG_DEBUG, " [DHCP_INIT] -> \n");
      } else if ( currState == &dhcpStop ) {
	syslog(LOG_DEBU, " [DHCP_STOP] -> \n");
      } else if ( currState == &dhcpStart ) {
	syslog(LOG_DEBUG, " [DHCP_START] -> \n");
      } else {
	syslog(LOG_DEBUG, " [DHCP_unknown] -> \n");
      }
#else	// DHCP_TIME_MODIFY
    //      syslog(LOG_INFO, "Transiting the state... [%d (%d)]",
    //	     time(NULL), dhcpUptime(NULL));
      if ( currState == &dhcpReboot ) {
	syslog(LOG_DEBUG, " [DHCP_REBOOT] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpInform ) {
	syslog(LOG_DEBUG, " [DHCP_INFORM] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpBound ) {
	syslog(LOG_DEBUG, " [DHCP_BOUND] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpRenew ) {
	syslog(LOG_DEBUG, " [DHCP_RENEW] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpRebind ) {
	syslog(LOG_DEBUG, " [DHCP_REBIND] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpInit ) {
	syslog(LOG_DEBUG, " [DHCP_INIT] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpStop ) {
	syslog(LOG_DEBUG, " [DHCP_STOP] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpStart ) {
	syslog(LOG_DEBUG, " [DHCP_START] -> (%d)\n", dhcpUptime(NULL));
      } else {
	syslog(LOG_DEBUG, " [DHCP_unknown] -> (%d)\n", dhcpUptime(NULL));
      }
#endif	// DHCP_TIME_MODIFY
    }
#endif	// DHCP_DEBUG_INFO

    if ( (currState=(void *(*)())currState()) == NULL ) exit(1);
  }
  while ( currState != &dhcpBound );

#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag ) {
#ifndef	DHCP_TIME_MODIFY
    syslog(LOG_DEBUG, "DHCP Bound. [%d]\n", time(NULL));
#else	// DHCP_TIME_MODIFY
    syslog(LOG_DEBUG, "DHCP Bound. [%d (%d)]\n", time(NULL), dhcpUptime(NULL));
#endif	// DHCP_TIME_MODIFY
  }
#endif	// DHCP_DEBUG_INFO

#ifndef DHCP_UPDATE
  if ( TestCase ) exit(0);
#endif /* !defined(DHCP_UPDATE) */

  alarm(0);
#ifdef DEBUG
  writePidFile(getpid());
#else

#ifdef  DHCP_UPDATE
#ifdef  EMBED
  s=vfork();
#else  /* !defined(EMBED) */
  s=fork();
#endif /* defined(EMBED) */
#else  /* !defined(DHCP_UPDATE) */
  s=fork();
#endif /* defined(DHCP_UPDATE) */

  if ( s )
    {

#ifdef	DHCP_DEBUG_INFO
      if ( DebugFlag ) {
	syslog(LOG_DEBUG, "Parent Process writing the pid...\n");
      }
#endif

      writePidFile(s);
      exit(0); /* got into bound state. */
    }
  setsid();

#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag ) {
    syslog(LOG_DEBUG, "Child became Parent and preparing logfiles... [%d]\n",
	   time(NULL));
  }
#endif

  if ( (i=open("/dev/null",O_RDWR,0)) >= 0 )
    {
      (void)dup2(i,STDIN_FILENO);
      (void)dup2(i,STDOUT_FILENO);
      (void)dup2(i,STDERR_FILENO);
      if ( i > 2 ) (void)close(i);
    }
#endif
  chdir("/");
  do {

#ifdef	DHCP_DEBUG_INFO
    if ( DebugFlag ) {
#ifndef	DHCP_TIME_MODIFY
      syslog(LOG_ERR, "Transiting the state... [%d]", time(NULL));
      if ( currState == &dhcpReboot ) {
	syslog(LOG_DEBUG, " [DHCP_REBOOT] -> \n");
      } else if ( currState == &dhcpInform ) {
	syslog(LOG_DEBUG, " [DHCP_INFORM] -> \n");
      } else if ( currState == &dhcpBound ) {
	syslog(LOG_DEBUG, " [DHCP_BOUND] -> \n");
      } else if ( currState == &dhcpRenew ) {
	syslog(LOG_DEBUG, " [DHCP_RENEW] -> \n");
      } else if ( currState == &dhcpRebind ) {
	syslog(LOG_DEBUG, " [DHCP_REBIND] -> \n");
      } else if ( currState == &dhcpInit ) {
	syslog(LOG_DEBUG, " [DHCP_INIT] -> \n");
      } else if ( currState == &dhcpStop ) {
	syslog(LOG_DEBUG, " [DHCP_STOP] -> \n");
      } else if ( currState == &dhcpStart ) {
	syslog(LOG_DEBUG, " [DHCP_START] -> \n");
      } else {
	syslog(LOG_DEBUG, " [DHCP_unknown] -> \n");
      }
#else
    //      syslog(LOG_INFO, "Transiting the state... [%d (%d)]",
    //	     time(NULL), dhcpUptime(NULL));
      if ( currState == &dhcpReboot ) {
	syslog(LOG_DEBUG, " [DHCP_REBOOT] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpInform ) {
	syslog(LOG_DEBUG, " [DHCP_INFORM] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpBound ) {
	syslog(LOG_DEBUG, " [DHCP_BOUND] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpRenew ) {
	syslog(LOG_DEBUG, " [DHCP_RENEW] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpRebind ) {
	syslog(LOG_DEBUG, " [DHCP_REBIND] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpInit ) {
	syslog(LOG_DEBUG, " [DHCP_INIT] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpStop ) {
	syslog(LOG_DEBUG, " [DHCP_STOP] -> (%d)\n", dhcpUptime(NULL));
      } else if ( currState == &dhcpStart ) {
	syslog(LOG_DEBUG, " [DHCP_START] -> (%d)\n", dhcpUptime(NULL));
      } else {
	syslog(LOG_DEBUG, " [DHCP_unknown] -> (%d)\n", dhcpUptime(NULL));
      }
#endif
    }
#endif	// DHCP_DEBUG_INFO

    currState=(void *(*)())currState();
  } while ( currState );
  deletePidFile();
  exit(1);
}
