/*
 * dhcpcd - DHCP client daemon -
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 * Copyright (C) September, 2002 Matsushita Electric Industrial Co., Ltd.
 *	Takahiko Suzuki <tsuzuki@vrl.mei.co.jp>
 * 
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include "udpipgen.h"

#ifdef	DHCP_DEBUG_INFO
#include <syslog.h>

extern	int		DebugFlag;
#endif

unsigned short ip_id;
/*****************************************************************************/
unsigned short in_cksum(addr,len)
unsigned short *addr;
int len;
{
  register int sum = 0;
  register u_short *w = addr;
  register int nleft = len;
  while ( nleft > 1 )
    {
      sum += *w++;
      nleft -= 2;
    }
  if ( nleft == 1 )
    {
      u_char a = 0;
      memcpy(&a,w,1);
      sum += a;
    }
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  return ~sum;
}
/*****************************************************************************/
void udpipgen(udpip,saddr,daddr,sport,dport,msglen)
udpiphdr *udpip;
unsigned int saddr,daddr;
unsigned short sport,dport,msglen;
{
  /* Use local copy because udpip->ip is not aligned.  */
  struct ip ip_local;
  struct ip *ip=&ip_local;
  struct ipovly *io=(struct ipovly *)udpip->ip;
  struct udphdr *udp=(struct udphdr *)udpip->udp;

#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "udpipgen entered. ");
#endif

  io->ih_next = io->ih_prev = 0;
  io->ih_x1 = 0;
  io->ih_pr = IPPROTO_UDP;
  io->ih_len = htons(msglen + sizeof(struct udphdr));
  io->ih_src.s_addr = saddr;
  io->ih_dst.s_addr = daddr;
  udp->uh_sport = sport;
  udp->uh_dport = dport;
  udp->uh_ulen = io->ih_len;
  udp->uh_sum = 0;
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum entering(1)... ");
#endif
  udp->uh_sum=in_cksum((unsigned short *)udpip,msglen+sizeof(udpiphdr));
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum returned. ");
#endif
  if ( udp->uh_sum == 0 ) udp->uh_sum = 0xffff;

#ifdef DHCP_UPDATE
  memcpy(ip,(struct ip *)udpip->ip,sizeof(ip_local));
#endif /* defined(DHCP_UPDATE) */

  ip->ip_hl = 5;
  ip->ip_v = IPVERSION;
  ip->ip_tos = 0;	/* normal service */
  ip->ip_len = htons(msglen + sizeof(udpiphdr));
  ip->ip_id = htons(ip_id++);
  ip->ip_off = 0;
  ip->ip_ttl = IPDEFTTL; /* time to live, 64 by default */
  ip->ip_p = IPPROTO_UDP;
  ip->ip_src.s_addr = saddr;
  ip->ip_dst.s_addr = daddr;
#ifdef __GLIBC__
  ip->ip_sum = 0;
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum entering(2)... ");
#endif

#ifdef  DHCP_UPDATE
  memcpy(udpip->ip,ip,sizeof(ip_local));
#endif /* defined(DHCP_UPDATE) */

  ip->ip_sum = in_cksum((unsigned short *)&ip_local,sizeof(struct ip));
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum returned. ");
#endif
#else
  ip->ip_csum = 0;
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum entering(2)... ");
#endif

#ifdef  DHCP_UPDATE
  memcpy(udpip->ip,ip,sizeof(ip_local));
#endif /* defined(DHCP_UPDATE) */

  ip->ip_csum = in_cksum((unsigned short *)&ip_local,sizeof(struct ip));
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum returned. ");
#endif
#endif
  memcpy(udpip->ip, ip, sizeof(struct ip));
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "udpipgen returns. ");
#endif
}
/*****************************************************************************/
int udpipchk(udpip)
udpiphdr *udpip;
{
  int hl;
  struct ip save_ip;
  struct ip *ip=(struct ip *)udpip->ip;
  struct ipovly *io=(struct ipovly *)udpip->ip;
  struct udphdr *udp=(struct udphdr *)udpip->udp;
  udpiphdr *nudpip = udpip;

#ifdef	DHCP_PACK_MODIFY
  char	zByte;
#endif

#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "udpipchk entered." );
#endif

#ifndef	DHCP_PACK_MODIFY
  hl = (ip->ip_hl)<<2;
#else
  zByte = *(char *)((struct ip *)(udpip->ip));
  hl = (zByte & 0xf) << 2;
#endif

#ifndef	DHCP_DEBUG_INFO
  if ( in_cksum((unsigned short *)udpip,hl) ) return -1;
#else
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum entering(3)... ");
  if ( in_cksum((unsigned short *)udpip,hl) )
  {
    syslog ( LOG_INFO, "*** ip_cksum error (ip). ***");
    return -1;
  }
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum returned ");
#endif
  memcpy(&save_ip, udpip->ip, sizeof(struct ip));
  hl -= sizeof(struct ip);
  if ( hl )
    { /* thrash IP options */
      nudpip = (udpiphdr *)((char *)udpip+hl);
      memmove((char *)nudpip,udpip,sizeof(struct ip));
      io=(struct ipovly *)nudpip->ip;
      ip=(struct ip *)nudpip->ip;
      udp=(struct udphdr *)nudpip->udp;
    }
#ifndef	DHCP_DEBUG_INFO
  if ( udp->uh_sum == 0 ) return 0; /* no checksum has been done by sender */
#else
  if ( udp->uh_sum == 0 )
  {
    if ( DebugFlag )
      syslog ( LOG_INFO, "udpipchk returns(1). (Sender doesn't implement) ");
    return 0; /* no checksum has been done by sender */
  }
#endif
  io->ih_next = io->ih_prev = 0;
  io->ih_x1 = 0;
  io->ih_len = udp->uh_ulen;
  hl = ntohs(udp->uh_ulen)+sizeof(struct ip);
#ifndef	DHCP_DEBUG_INFO
  if ( in_cksum((unsigned short *)nudpip,hl) ) return -2;
#else
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum entering(4)... ");
  if ( in_cksum((unsigned short *)nudpip,hl) )
  {
    syslog ( LOG_INFO, "*** in_cksum error (udp). ***");
    return -2;
  }
  if ( DebugFlag )
    syslog ( LOG_INFO, "in_cksum returned ");
#endif
  memcpy(udpip->ip, &save_ip, sizeof(struct ip));
#ifdef	DHCP_DEBUG_INFO
  if ( DebugFlag )
    syslog ( LOG_INFO, "udpipchk returns(2). ");
#endif
  return 0;
}
